"""
Relevance search REST API.

This module provides the REST API for search.
"""

import os
import pydoc
import anyconfig

from flask import request
from flask import jsonify
from flask import Response
from werkzeug import exceptions

from relevance.api import Flask
from relevance.api import start_app
from relevance.query import QueryParser
from relevance.query import QueryParserError
from relevance.facet import Facet
from relevance.engine import SearchEngine
from relevance.engine import MappingEngine
from relevance.engine import EngineError
from relevance.engine import SearchError
from relevance.engine import MappingError


# Register the application
app = Flask('relevance.api.search')

# The configuration objects
app.data = type('app_data', (object,), dict(
    config={},
    engines={},
))()


def load_config(env_name: str=None):
    """
    Load or reload the configuration.

    :param env_name: the name of the environment to load. If omitted, it is taken
    from the PYTHON_ENV environment variable. If it doesn't exist, 'development' is
    assumed.
    """
    if env_name is None:
        env_name = os.environ.get('PYTHON_ENV', 'development')

    # Load the configuration file
    this_config = anyconfig.load(
        ['./etc/search.json', './etc/search.{0}.json'.format(env_name)],
        ignore_missing=True,
    )

    # Load the engine instances
    this_items = {}
    for name, conf in this_config.get('engines', {}).items():
        try:
            cls_name = conf['engine']
            target = conf['target']
        except KeyError as ex:
            raise KeyError('missing configuration key {0}'.format(str(ex)))

        cls = pydoc.locate(cls_name)
        if cls is None or not issubclass(cls, SearchEngine):
            raise ImportError('cannot find engine {0}'.format(cls_name))

        facets = {}
        for facet_name, facet_conf in conf.get('facets', {}).items():
            facet_cls = pydoc.locate(facet_conf.get('type'))
            if facet_cls is None or not issubclass(facet_cls, Facet):
                raise ImportError('cannot find facet {0}'.format(facet_conf['type']))

            facet_obj = facet_cls(
                field=facet_conf.get('field'),
                path=facet_conf.get('path'),
                **facet_conf.get('options', {}),
            )

            facets[facet_name] = facet_obj

        options = conf.get('options', {})
        this_items[name] = cls(target, facets=facets, **options)

    app.data.config.clear()
    app.data.config.update(this_config)
    app.data.engines.clear()
    app.data.engines.update(this_items)


def load_plugins():
    """
    Load or reload the plugins.
    """
    for plugin in app.data.config.get('plugins', []):
        func = pydoc.locate(plugin)
        if func is None:
            raise ImportError('cannot load plugin {0}'.format(plugin))

        func(app)


@app.route('/', methods=['GET'])
def get_config() -> Response:
    """
    Get the configuration.
    """
    response = jsonify(app.data.config)
    response.status_code = 201
    return response


@app.route('/<engine_name>', methods=['GET'])
def do_search(engine_name: str) -> Response:
    """
    Perform a search request.

    Request arguments:
        q -- the query to run.

    :param engine_name: the name of the engine to use.
    :returns: a result set object.
    """
    try:
        engine = app.data.engines[engine_name]
    except KeyError:
        raise EngineNotFoundError(engine_name)

    query = request.args.get('q', '')
    search = QueryParser().loads(query)

    results = engine.search(search)
    data = {
        'search': search.to_dict(),
        'results': results,
        'time': results.time,
        'count': results.count,
    }

    if results.facets is not None:
        data.update({
            'facets': results.facets,
        })

    response = jsonify(data)
    response.status_code = 200
    return response


@app.route('/<engine_name>/mapping', methods=['GET'])
def get_mapping_doc_types(engine_name: str) -> Response:
    """
    Get a list of available doc types.

    :param engine_name: the name of the engine to use.
    :returns: a list of document types.
    """
    try:
        engine = app.data.engines[engine_name]

        if not isinstance(engine, MappingEngine):
            raise MappingNotSupportedError(engine_name)
    except KeyError:
        raise EngineNotFoundError(engine_name)

    data = engine.get_doc_types()
    response = jsonify(data)
    response.status_code = 200
    return response


@app.route('/<engine_name>/mapping/<doc_type>', methods=['GET'])
def get_mapping(engine_name: str, doc_type: str) -> Response:
    """
    Get a list of available doc types.

    :param engine_name: the name of the engine to use.
    :returns: a list of document types.
    """
    try:
        engine = app.data.engines[engine_name]

        if not isinstance(engine, MappingEngine):
            raise MappingNotSupportedError(engine_name)

    except KeyError:
        raise EngineNotFoundError(engine_name)

    data = engine.get_mapping(doc_type).to_dict()
    response = jsonify(data)
    response.status_code = 200
    return response


@app.after_request
def app_cors(response: Response) -> Response:
    """
    CORS request handler.

    Adds the CORS headers to all responses so that the API can be used directly from
    a web browser.
    """
    headers = {
        'Access-Control-Allow-Origin': request.headers.get('Origin', '*'),
        'Access-Control-Allow-Credentials': 'true',
        'Access-Control-Allow-Methods': 'POST, OPTIONS, GET, PUT, DELETE, HEAD, TRACE',
        'Access-Control-Allow-Headers': request.headers.get(
            'Access-Control-Request-Headers', 'Authorization'
        ),
    }

    for k, v in headers.items():
        response.headers[k] = v

    if app.debug:
        response.headers['Access-Control-Max-Age'] = '1'

    return response


class EngineNotFoundError(exceptions.NotFound):
    """
    Exception raised when a requested engine is not found.
    """
    pass


@app.errorhandler(EngineNotFoundError)
def engine_not_found_error(e: EngineNotFoundError) -> Response:
    response = jsonify({
        'error': {
            'desc': 'engine is not configured',
            'key': str(e),
            'code': e.__class__.__name__,
        }
    })
    response.status_code = 404
    return response


class MappingNotSupportedError(exceptions.NotImplemented):
    """
    Exception raised when requesting a mapping from an unsupported engine.
    """
    pass


@app.errorhandler(MappingNotSupportedError)
def mapping_not_supported_error(e: MappingNotSupportedError) -> Response:
    response = jsonify({
        'error': {
            'desc': 'engine does not support mapping',
            'key': str(e),
            'code': e.__class__.__name__,
        }
    })
    response.status_code = 501
    return response


@app.errorhandler(QueryParserError)
def query_parser_error(e: QueryParserError) -> Response:
    """
    Handle parse errors.
    """
    response = jsonify({
        'error': {
            'desc': str(e),
            'code': e.__class__.__name__,
        }
    })
    response.status_code = 400
    return response


@app.errorhandler(EngineError)
def engine_error(e: EngineError) -> Response:
    response = jsonify({
        'error': {
            'desc': str(e),
            'code': e.__class__.__name__,
        }
    })
    response.status_code = 503
    return response


@app.errorhandler(SearchError)
def search_error(e: SearchError) -> Response:
    response = jsonify({
        'error': {
            'desc': str(e),
            'code': e.__class__.__name__,
        }
    })
    response.status_code = 503
    return response


@app.errorhandler(MappingError)
def mapping_error(e: MappingError) -> Response:
    response = jsonify({
        'error': {
            'desc': str(e),
            'code': e.__class__.__name__,
        }
    })
    response.status_code = 503
    return response


@app.errorhandler(400)
@app.errorhandler(404)
@app.errorhandler(405)
@app.errorhandler(500)
@app.errorhandler(exceptions.HTTPException)
def http_error(e: exceptions.HTTPException):
    response = jsonify({
        'error': {
            'desc': '{0}: {1}'.format(e.__class__.__name__, str(e)),
            'code': 'HTTPError',
        }
    })
    response.status_code = 500
    return response


@app.errorhandler(Exception)
def exception_error(e: Exception) -> Response:
    response = jsonify({
        'error': {
            'desc': '{0}: {1}'.format(e.__class__.__name__, str(e)),
            'code': 'Exception',
        }
    })
    response.status_code = 500
    return response


if __name__ == '__main__':
    load_config()
    load_plugins()
    start_app(app)



"""
Relevance pipeline REST module.

This module provides an interface for dealing with queues via an API.
"""

import json
import pickle
import typing
from flask import request
from flask import jsonify
from flask import Response
from queue import Queue
from queue import Empty
from werkzeug import exceptions

from relevance.pipeline import EmptyError
from relevance.pipeline import Extractor
from relevance.api.pipelines import app
from relevance.api.pipelines import PipelineNotFoundError


class QueueExtractor(Queue, Extractor):
    """
    Queue extractor.

    This class provides an extractor that wraps a queue that can be populated via
    an API.
    """
    def get(self, *args, **kwargs) -> object:
        """
        Get an item from the extractor.

        :raises: EmptyError: if the queue is empty.
        """
        try:
            return super().get(*args, **kwargs)
        except Empty:
            raise EmptyError()

    def get_status(self) -> typing.Dict[str, object]:
        """
        Get additional state for the extractor.
        """
        return {
            'size': self.qsize(),
        }


@app.route('/<pipeline_name>', methods=['PUT'])
def put_document(pipeline_name: str) -> Response:
    """
    Put a document in the queue.

    Request payload:
        content -- the content to parse.
        type -- a MIME type of parse the content as. When None, the content is assumed
        to be a valid JSON object that can be fed directly into a document object.

    :returns: a 202 response on success.
    """
    try:
        pipeline = app.data.pipelines[pipeline_name]
    except KeyError:
        raise PipelineNotFoundError(pipeline_name)

    if not isinstance(pipeline, QueueExtractor):
        raise exceptions.MethodNotAllowed()

    payload = request.get_json()
    mime_type = payload.get('type')

    try:
        content = payload['content']
    except KeyError as e:
        raise MissingParameterError(str(e))

    if mime_type is not None:
        if mime_type == 'application/json':
            content = json.loads(content)
        elif mime_type == 'application/octet-stream':
            content = pickle.loads(content)
        else:
            raise DocumentTypeError(mime_type)

    pipeline.extractor.put(content)

    response = Response()
    response.status_code = 202
    return response


class DocumentTypeError(exceptions.UnprocessableEntity):
    """
    Exception raised when a document type cannot be decoded.
    """
    pass


@app.errorhandler(DocumentTypeError)
def document_type_error(e: DocumentTypeError) -> Response:
    response = jsonify({
        'error': {
            'desc': 'the supplied document type cannot be decoded',
            'key': str(e),
            'code': e.__class__.__name__,
        }
    })
    response.status_code = 501
    return response


class MissingParameterError(exceptions.BadRequest):
    """
    Exception raised when a parameter is missing from the payload.
    """
    pass


@app.errorhandler(MissingParameterError)
def missing_param_error(e: MissingParameterError) -> Response:
    response = jsonify({
        'error': {
            'desc': 'the payload is missing a required parameter',
            'key': str(e),
            'code': e.__class__.__name__,
        }
    })
    response.status_code = 501
    return response



"""
Relevance ingestion pipeline.

This package provides the classes and interfaces for ingesting content using the
engine abstraction layer.
"""

import abc
import typing
import logging
import threading

from relevance.document import Document


# Logging
logger = logging.getLogger('relevance.pipeline')


class EmptyError(IOError):
    """
    Exception raised when an extractor encounters an error.
    """
    pass


class Extractor(object, metaclass=abc.ABCMeta):
    """
    Extractor abstract class.

    An extractor's responsibility is to pull raw content into the pipeline for
    further processing.
    """
    @abc.abstractmethod
    def get(self, timeout: int) -> object:
        """
        Get an item from the extractor.

        :param timeout: a timeout, in seconds after which an exception is raised
        if an item cannot be retrieved.
        :returns: the retrieved item.
        """
        pass

    @abc.abstractmethod
    def task_done(self):
        """
        Let the extractor know that an item was processed.

        This method is called after each item is processed.
        """
        pass

    @abc.abstractproperty
    def get_status(self) -> typing.Dict[str, object]:
        """
        Get additional state for the extractor.
        """
        pass


class Publisher(object):
    """
    Publisher class.

    A publisher is an object that takes processed content and publishes to
    a data engine.
    """
    def __init__(self, func: typing.Callable):
        """
        Initialize the publisher.

        :param func: a function to call when publishing.
        """
        self.func = func
        self.publish_count = 0
        self.error_count = 0

        logger.debug('initialize publisher {0} with {1}'.format(
            self, self.func,
        ))

    def publish(self, content: object) -> object:
        """
        Publish a data item.

        :param content: the content to publish.
        :returns: returns a status item.
        """
        try:
            result = self.func(content)
            self.publish_count += 1
            logger.info('publisher {0} item published: {1}'.format(
                self, content,
            ))
            return result
        except Exception as e:
            self.error_count += 1
            logger.error('publisher {0} item error: {1}'.format(
                self, e,
            ))
            raise

    def get_status(self) -> typing.Dict[str, object]:
        """
        Get additional state for the publisher.
        """
        return {
            'publish_count': self.publish_count,
            'error_count': self.error_count,
        }


class DocumentExtractor(Extractor):
    """
    Extracts a dictionary payload into a document.

    This class wraps around an existing extractor and converts the input dictionary into a
    document object.
    """
    def __init__(self, handler: Extractor, schema: str=None, doc_type: str=None):
        """
        Initialize the extractor.

        :param handler: the actual extractor handler to launch after the document is converted.
        :param schema: the name of the schema to apply to the resulting document.
        :param doc_type: the document type to apply to the resulting document.
        """
        self.handler = handler
        self.schema = schema
        self.doc_type = doc_type

        logger.debug('initialize document extractor {0} with {1} to {2} as {3}'.format(
            self, self.handler, self.schema, self.doc_type,
        ))

    def get(self, timeout: int) -> object:
        """
        Get an item from the extractor.

        :raises: KeyError: when a required payload element is missing.
        """
        item = self.handler.get(timeout)

        try:
            schema = self.schema if self.schema is not None else item['_schema']
            doc_type = self.doc_type if self.doc_type is not None else item['_type']
            uid = item.get('_uid', None)
        except KeyError as e:
            raise KeyError('missing payload item {0}'.format(str(e)))

        doc = Document(schema, doc_type, uid)
        doc.update(item)

        return doc

    def task_done(self):
        """
        Let the extractor know that an item was processed.
        """
        return self.handler.task_done()

    def __getattr__(self, name: str) -> object:
        """
        Overload get attribute method.
        """
        return getattr(self.handler, name)


class Pipeline(object):
    """
    Ingestion pipeline class.

    The class encapsulates an extractor and a list of callables to execute for each
    input item.
    """
    def __init__(self, extractor: Extractor, publisher: Publisher, *args):
        """
        Initialize the pipeline.

        :param extractor: a extractor object containing the input data.
        :param publisher: a publisher object to send the output data.
        :param args: a list of tuples or callables to add to the pipeline.
        """
        self.extractor = extractor
        self.publisher = publisher
        self.funcs = []

        for x in args:
            if isinstance(x, tuple):
                self.funcs.append(x[0], x[1])
            else:
                self.funcs.append((x, None))

        logger.debug('initialize pipeline {0}, ext={1}, pub={2}, func={3}'.format(
            self, self.extractor, self.publisher, self.funcs,
        ))

    def run(self, timeout: int=None) -> object:
        """
        Run an item from the extractor.

        :param timeout: a timeout to wait before an item becomes available.
        :returns: the processed item.
        """
        try:
            item = self.extractor.get(timeout=timeout)
        except EmptyError as e:
            raise

        logger.debug('pipeline {0} extracted item: {1}'.format(
            self, item,
        ))

        try:
            for func, desc in self.funcs:
                logger.debug('pipeline {0} applying {1} ({2}): {3}'.format(
                    self, func, desc, item,
                ))

                item = func(item)

            logger.debug('pipeline {0} publishing: {1}'.format(
                self, item,
            ))

            return self.publisher(item)
        finally:
            self.extractor.task_done()
            logger.debug('pipeline {0} release item'.format(
                self,
            ))


class ThreadedPipeline(Pipeline):
    """
    Threaded pipeline class.

    A pipeline variant that runs in a separate thread.
    """
    def __init__(self, *args, **kwargs):
        """
        Initialize the pipeline.
        """
        super().__init__(*args, **kwargs)
        self.thread = None
        self.is_started = False

    def start(self, timeout: float=1):
        """
        Start the pipeline thread.

        :param timeout: the amount of time, in seconds, to wait for the extractor at
        in the event loop.
        """
        if self.thread is not None:
            logger.warn('threaded pipeline {0} already started'.format(
                self,
            ))
            return

        def run_thread():
            logger.debug('threaded pipeline {0} thread starting'.format(
                self,
            ))

            while True:
                try:
                    result = self.run(timeout=timeout)
                    logger.info('threaded pipeline {0} item processed: {1}'.format(
                        self, result,
                    ))
                except EmptyError:
                    if not self.is_started:
                        break
                except StopIteration as e:
                    logger.info('threaded pipeline {0} skip item: {1}'.format(
                        self, e,
                    ))
                except Exception as e:
                    logger.error('threaded pipeline {0} item error: {1}'.format(
                        self, e,
                    ))

            logger.debug('threaded pipeline {0} thread ending'.format(
                self,
            ))

        self.is_started = True
        self.thread = threading.Thread(target=run_thread)
        self.thread.start()

        logger.info('threaded pipeline {0} starting {1}'.format(
            self, self.thread,
        ))

    def stop(self):
        """
        Stop the pipeline thread.
        """
        if self.thread is None:
            logger.warn('threaded pipeline {0} already stopped'.format(
                self,
            ))
            return

        self.is_started = False
        self.thread.join()

        logger.info('threaded pipeline {0} joined'.format(
            self,
        ))



"""
Relevance pipeline spider module.

This module provides an interface for ingesting content by crawling web pages.
"""

import typing
import logging
from queue import Queue
from scrapy import Spider
from scrapy import signals
from scrapy.crawler import Crawler
from scrapy.crawler import CrawlerRunner
from twisted.internet import reactor

from relevance.pipeline import Extractor


# Logging
logger = logging.getLogger('relevance.pipeline.spider')


class SpiderExtractor(Extractor):
    """
    Spider extractor.

    This classes wraps a scrapy spider into a pipeline extractor.
    """
    # The number of running instances
    __instances = Queue()

    def __init__(self, spider: Spider):
        """
        Initialize the extractor.

        :param spider: the spider object to use.
        """
        self.spider = spider
        self.queue = Queue()
        self.urls = set()

        def add_item(item, response, sender, spider, signal):
            if response.url not in self.urls:
                logger.info('spider extractor {0} extract item {1}: {2}'.format(
                    self, item, response,
                ))

                self.queue.put(item)
                self.urls.add(response.url)

        def start_spider():
            self.__instances.put(self)

        def stop_spider():
            self.__instances.get_nowait()
            if self.__instances.empty():
                reactor.stop()

        self.crawler = Crawler(spider)
        self.crawler.signals.connect(add_item, signals.item_scraped)
        self.crawler.signals.connect(start_spider, signals.spider_opened)
        self.crawler.signals.connect(stop_spider, signals.spider_closed)
        self.runner = CrawlerRunner()
        self.runner.crawl(self.crawler)

        logger.debug('initialize spider extractor {0}'.format(
            self,
        ))

        if reactor._stopped:
            reactor.run()

            logger.debug('spider extractor {0} starting reactor'.format(
                self,
            ))

    def get(self, timeout: int) -> object:
        """
        Get an item from the extractor.
        """
        return self.queue.get(timeout)

    def task_done(self):
        """
        Let the extractor know that an item was processed.
        """
        return self.queue.task_done()

    def get_status(self) -> dict:
        """
        Get additional state for the extractor.
        """
        return {
            'is_running': reactor._started,
            'size': self.queue.qsize(),
            'urls': self.spider.start_urls
        }


class SimpleSpiderExtractor(SpiderExtractor):
    """
    Simple spider extractor implementation.

    This classes automatically configures a spider extractor to extract data based
    on a certain mapping and follow URLs that match a certain selector.
    """
    def __init__(self, urls: typing.List[str], mapping: typing.Dict[str, str],
                 follow: typing.List[str]=None, match_urls: typing.List[str]=None):
        """
        Initialize the extractor.

        :param urls: the URLs to crawl.
        :param mapping: a dictionary mapping field names (keys) as scrapy selectors.
        :param follow: a list of selectors to check for links to follow.
        :param match_urls: a list of regular expressions to match against before following a
        link. If any matches, the link will be followed.
        """
        self.urls = urls
        self.mapping = mapping
        self.follow = follow or ['a']
        self.match_urls = match_urls or []

        class SimpleSpider(Spider):
            name = 'simplespider'
            start_urls = urls

            def parse(this, response):
                for selector in self.follow:
                    for next_page in response.css(selector):
                        href = next_page.css('::attr(href)')
                        found = False
                        for x in self.match_urls:
                            if href.re(x):
                                found = True

                        if found:
                            yield response.follow(next_page, this.parse)

                result = {}
                for field, selector in self.mapping.items():
                    result[field] = response.css(selector).extract()
                yield result

        super().__init__(SimpleSpider)



"""
Relevance facets module.

This module provides the interface definitions for handling facets.
"""

import abc
import dateutil.parser
from dateutil.relativedelta import relativedelta


class FacetDefinitionError(TypeError):
    """
    Exception class raised when a facet is not defined properly.
    """
    pass


class FacetValueError(ValueError):
    """
    Exception class raised when a requested facet value is invalid.
    """
    def __init__(self, value: str):
        """
        Initialize the exception.

        :param value: the value of the facet that caused the error.
        """
        super().__init__('invalid facet value {0}'.format(value))
        self.value = value


class FacetNotDefinedError(NameError):
    """
    Exception class raised when a requested facet is not defined.
    """
    def __init__(self, name: str):
        """
        Initialize the exception.

        :param name: the name of the facet that caused the error.
        """
        super().__init__('the facet {0} is not defined'.format(name))
        self.name = name


class FacetNotSupportedWarning(UserWarning):
    """
    Warning class when an engine does not support a specifi facet.
    """
    def __init__(self, cls: type):
        """
        Initialize the exception.

        :param cls: the facet class that raised the warning.
        """
        super().__init__('the facet type {0} is not supported by the engine'.format(
            cls.__name__
        ))
        self.cls = cls


class Facet(object):
    """
    Base facet definition class.
    """
    def __init__(self, *, field: str=None, path: str=None):
        """
        Initialize the definition.

        :param field: the field for the facet.
        :param path: the path or the facet. Mutually exclusive with field. Used when
        a facet is nested.
        :raises FacetDefinitionError: if the facet is defined improperly.
        """
        if field is not None and path is not None:
            raise FacetDefinitionError('field and path are mutually exclusive')

        self.field = field
        self.path = path


class TermFacet(Facet):
    """
    Facet type for term facets.

    Aggregates on matching terms.
    """
    pass


class HistogramFacet(Facet, metaclass=abc.ABCMeta):
    """
    Facet type for histogram facets.

    Facets that use ranges should inherit from this abstract class.
    """
    @abc.abstractmethod
    def get_range_for(self, value: object) -> tuple:
        """
        Get a start and end range given a specific value.

        :param value: the value to get the range for.
        :returns: a tuple containing the start and end values for the range.
        """
        pass


class DateFacet(HistogramFacet):
    """
    Facet type for date/time histogram facets.

    Aggregates on date intervals.
    """
    # Interval definitions
    YEAR = 'year'
    QUARTER = 'quarter'
    MONTH = 'month'
    WEEK = 'week'
    DAY = 'day'
    HOUR = 'hour'
    MINUTE = 'minute'
    SECOND = 'second'

    def __init__(self, *, interval: str=MONTH, **kwargs):
        """
        Initialize the facet.

        :param interval: the interval at which to group the results.
        """
        super().__init__(**kwargs)
        self.interval = interval

    def get_range_for(self, value: object) -> tuple:
        """
        Override abstract method.

        :raises FacetValueError: if an invalid range is specified.
        :raises FacetDefinitionError: if the facet is defined improperly.
        """
        start = dateutil.parser.parse(value)
        prop, delta = self.interval, 1

        if self.interval == DateFacet.YEAR:
            if start.month != 1 or start.day != 1 or start.hour != 0 or \
               start.minute != 0 or start.second != 0:
                raise FacetValueError(value)

        elif self.interval == DateFacet.QUARTER:
            if start.month not in [1, 4, 7, 10] or start.day != 1 or \
               start.hour != 0 or start.minute != 0 or start.second != 0:
                raise FacetValueError(value)
            prop, delta = 'month', 3

        elif self.interval == DateFacet.MONTH:
            if start.day != 1 or start.hour != 0 or \
               start.minute != 0 or start.second != 0:
                raise FacetValueError(value)

        elif self.interval == DateFacet.WEEK:
            if start.weekday() != 0:
                raise FacetValueError(value)
            prop, delta = 'day', 7

        elif self.interval == DateFacet.DAY:
            if start.hour != 0 or start.minute != 0 or start.second != 0:
                raise FacetValueError(value)

        elif self.interval == DateFacet.HOUR:
            if start.minute != 0 or start.second != 0:
                raise FacetValueError(value)

        elif self.interval == DateFacet.MINUTE:
            if start.second != 0:
                raise FacetValueError(value)

        else:
            raise FacetDefinitionError('invalid interval {0}'.format(self.interval))

        end = start + relativedelta(**{'{0}s'.format(prop): delta})

        return start.isoformat(), end.isoformat()


class IntervalFacet(HistogramFacet):
    """
    Facet type for numeric interval facets.

    Aggregates on numeric intervals.
    """
    def __init__(self, *, interval: object, **kwargs):
        """
        Initialize the facet.

        :param interval: the interval at which to group the results.
        """
        super().__init__(**kwargs)
        self.interval = interval

    def get_range_for(self, value: object) -> tuple:
        """
        Override abstract method.

        :raises FacetValueError: if an invalid range is specified.
        """
        if (value % self.interval) > 0:
            raise FacetValueError(self, value)

        return value, value + self.interval


class RangeFacet(HistogramFacet):
    """
    Facet type for rangeed histogram facets.

    Aggregates on specific numeric intervals.
    """
    def __init__(self, *, ranges: dict, **kwargs):
        """
        Initialize the facet.

        :param ranges: a dictionary of ranges with each value being a tuple containing
        a start and an end value.
        """
        super().__init__(**kwargs)
        self.ranges = ranges

    def get_range_for(self, value: object) -> tuple:
        """
        Override abstract method.

        :raises FacetValueError: if an invalid range is specified.
        """
        try:
            start, end = self.ranges[value]
        except KeyError:
            raise FacetValueError(self, value)

        return start, end



"""
Content pipeline manager.

Relevance is a software suite that aims to address the common and recurring problem
of ingesting content, delivering it, searching it with relevant results and analyzing
its usage by the end-users, acting on the analytic data to provide the best results
possible.
"""

# Package version
__version__ = '0.3.1'


def init_app():
    """
    Initialize the application.
    """
    import os
    import sys

    base_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

    for path in ['usr']:
        full_path = '{0}/{1}'.format(base_path, path)
        if full_path not in sys.path:
            sys.path.append(full_path)


# Bootstrap
init_app()



"""
Relevance pipeline REST API.

This module provides the REST API for pipelines.
"""

import os
import pydoc
import anyconfig

from flask import request
from flask import jsonify
from flask import Response
from werkzeug import exceptions

from relevance.api import Flask
from relevance.api import start_app
from relevance.pipeline import Pipeline


# Register the application
app = Flask('relevance.api.pipelines')

# The configuration objects
app.data = type('app_data', (object,), dict(
    config={},
    pipelines={},
))()


def load_config(env_name: str=None):
    """
    Load or reload the configuration.

    :param env_name: the name of the environment to load. If omitted, it is taken
    from the PYTHON_ENV environment variable. If it doesn't exist, 'development' is
    assumed.
    """
    if env_name is None:
        env_name = os.environ.get('PYTHON_ENV', 'development')

    # Load the configuration file
    this_config = anyconfig.load(
        ['./etc/pipelines.json', './etc/pipelines.{0}.json'.format(env_name)],
        ignore_missing=True,
    )

    # Load the pipeline instances
    this_items = {}
    for name, target in this_config.get('pipelines', {}).items():
        loader = pydoc.locate(target)

        pipe = loader()
        if not isinstance(pipe, Pipeline):
            raise ImportError('cannot load pipeline from {0}'.format(target))

        this_items[name] = pipe

    app.data.config.clear()
    app.data.config.update(this_config)
    app.data.pipelines.clear()
    app.data.pipelines.update(this_items)


def load_plugins():
    """
    Load or reload the plugins.
    """
    for plugin in app.data.config.get('plugins', []):
        func = pydoc.locate(plugin)
        if func is None:
            raise ImportError('cannot load plugin {0}'.format(plugin))

        func(app)


@app.route('/', methods=['GET'])
def get_config() -> Response:
    """
    Get the configuration.
    """
    data = list(app.data.config.keys())
    response = jsonify(data)
    response.status_code = 201
    return response


@app.route('/<pipeline_name>', methods=['GET'])
def get_pipeline_info(pipeline_name: str) -> Response:
    """
    Get information about a pipeline.

    :param pipeline_name: the name of the pipeline to check.
    :returns: a status object.
    """
    try:
        pipeline = app.data.pipelines[pipeline_name]
    except KeyError:
        raise PipelineNotFoundError(pipeline_name)

    try:
        input_status = pipeline.extractor.input_status or {}
    except AttributeError:
        input_status = {'size': pipeline.extractor.qsize()}

    try:
        output_status = pipeline.publisher.output_status or {}
    except AttributeError:
        output_status = {}

    data = {
        'type': '{0}.{1}'.format(pipeline.__class__.__module__, pipeline.__class__.__name__),
        'target': app.data.config[pipeline_name],
        'input': input_status,
        'output': output_status,
    }

    response = jsonify(data)
    response.status_code = 200
    return response


@app.after_request
def app_cors(response: Response) -> Response:
    """
    CORS request handler.

    Adds the CORS headers to all responses so that the API can be used directly from
    a web browser.
    """
    headers = {
        'Access-Control-Allow-Origin': request.headers.get('Origin', '*'),
        'Access-Control-Allow-Credentials': 'true',
        'Access-Control-Allow-Methods': 'POST, OPTIONS, GET, PUT, DELETE, HEAD, TRACE',
        'Access-Control-Allow-Headers': request.headers.get(
            'Access-Control-Request-Headers', 'Authorization'
        ),
    }

    for k, v in headers.items():
        response.headers[k] = v

    if app.debug:
        response.headers['Access-Control-Max-Age'] = '1'

    return response


class PipelineNotFoundError(exceptions.NotFound):
    """
    Exception raised when a requested pipeline is not found.
    """
    pass


@app.errorhandler(PipelineNotFoundError)
def pipeline_not_found_error(e: PipelineNotFoundError) -> Response:
    response = jsonify({
        'error': {
            'desc': 'pipeline is not configured',
            'key': str(e),
            'code': e.__class__.__name__,
        }
    })
    response.status_code = 404
    return response


@app.errorhandler(400)
@app.errorhandler(404)
@app.errorhandler(405)
@app.errorhandler(500)
@app.errorhandler(exceptions.HTTPException)
def http_error(e: exceptions.HTTPException):
    response = jsonify({
        'error': {
            'desc': '{0}: {1}'.format(e.__class__.__name__, str(e)),
            'code': 'HTTPError',
        }
    })
    response.status_code = 500
    return response


@app.errorhandler(Exception)
def exception_error(e: Exception) -> Response:
    response = jsonify({
        'error': {
            'desc': '{0}: {1}'.format(e.__class__.__name__, str(e)),
            'code': 'Exception',
        }
    })
    response.status_code = 500
    return response


if __name__ == '__main__':
    load_config()
    load_plugins()
    start_app(app)



"""
Relevance engine core.

This package provides a unified interface for interacting with the different storage
and search backends.
"""

import abc
import typing
import logging

from relevance.query import Search
from relevance.facet import Facet
from relevance.mapping import Mapping
from relevance.document import Document
from relevance.document import ResultSet


# Logging
logger = logging.getLogger('relevance.engine')


class EngineError(RuntimeError):
    """
    Exception class for engine errors.
    """
    pass


class MappingError(RuntimeError):
    """
    Exception class for mapping errors.
    """
    pass


class MappingEngine(object, metaclass=abc.ABCMeta):
    """
    Abstract class for engines that support mapping.

    This class provides the interface for storage and search engines to support
    mapping definitions.
    """
    def __init__(self, target: str, **kwargs):
        """
        Initialize the engine.

        :param target: the storage engine target.
        """
        logger.info('initialize mapping engine {0} at {1}'.format(
            self, target,
        ))
        self.target = target

    @abc.abstractmethod
    def get_doc_types(self) -> typing.List[str]:
        """
        Get a list of available document types.

        :returns: a list of document type strings.
        """
        pass

    @abc.abstractmethod
    def get_mapping(self, doc_type: str) -> Mapping:
        """
        Get the mapping object for a specific document type.

        :param doc_type: the document type.
        :returns: the mapping object.
        """
        pass

    @abc.abstractmethod
    def put_mapping(self, doc_type: str, mapping: Mapping):
        """
        Update the mapping for a specific document type.

        :param doc_type: the document type.
        :param mapping: the mapping object.
        """
        pass


class SearchError(RuntimeError):
    """
    Exception class for mapping errors.
    """
    pass


class SearchEngine(object, metaclass=abc.ABCMeta):
    """
    Search Engine abstract class.

    This class provides the abstract interface for storage engines to implement search
    functionality.
    """
    def __init__(self, target: str, *, facets: typing.Dict[str, Facet]=None,
                 auto_filters: bool=True, **kwargs):
        """
        Initialize the engine.

        :param target: the storage engine target.
        :param facets: a dictionary of facets definitions.
        :param auto_filters: whether to enable automatic filter definitions.
        """
        logger.info('initialize search engine {0} at {1}, auto_filters={2}, facets={3}'.format(
            self, target, int(auto_filters), facets,
        ))
        self.target = target
        self.facets = facets if facets is not None else {}
        self.auto_filters = auto_filters

    @abc.abstractmethod
    def search(self, search: Search) -> ResultSet:
        """
        Perform a search request on the engine.

        :param search: the request to execute.
        :returns: a result set object.
        """
        pass


class IngestionError(RuntimeError):
    """
    Exception class for ingestion errors.
    """
    pass


class IngestionEngine(object, metaclass=abc.ABCMeta):
    """
    Ingestion Engine abstract class.

    This class provides the abstract interface for storage engines to implement indexing
    functionality.
    """
    def __init__(self, target: str, **kwargs):
        """
        Initialize the engine.

        :param target: the storage engine target.
        """
        logger.info('initialize ingestion engine {0} at {1}'.format(
            self, target,
        ))
        self.target = target

    @abc.abstractmethod
    def publish(self, doc: Document) -> Document:
        """
        Index a document.

        :param doc: the document object to index.
        :returns: the updated, indexed document object.
        """
        pass

    @abc.abstractmethod
    def unpublish(self, schema: str, doc_type: str, uid: object):
        """
        Delete a document.

        :param schema: the document schema.
        :param doc_type: the document type.
        :param uid: the unique identifier for the document.
        """
        pass



"""
Relevance API package.

This package provides some common interfaces for the different REST APIs to use.
"""

import logging
from flask import Flask
from flask import request


def start_app(app: Flask):
    """
    Start an API application.

    :param app: the application object.
    """
    init_logger(app)
    app.run()


def init_logger(app: Flask):
    """
    Enable the API logger.

    :param app: the application object.
    """
    logger = logging.getLogger(app.name)

    @app.before_request
    def logger_before():
        logger.info('received {0} "{1}" with {2} bytes of {3}'.format(
            request.method, request.full_path, len(request.data),
            request.headers.get('Content-Type', 'unknown'),
        ))
        logger.debug('received {0} "{1}": {2}'.format(
            request.method, request.url, request.data,
        ))

    @app.after_request
    def logger_after(response):
        logger.info('response {0} "{1}" {2} with {3} bytes of {4}'.format(
            request.method, request.full_path, response.status_code, len(response.data),
            response.headers.get('Content-Type', 'unknown'),
        ))
        logger.debug('response {0} "{1}" {2}: {3}'.format(
            request.method, request.url, response.status_code, response.data,
        ))
        return response
